﻿using UnityEngine;

public class Timer
{
    public float Duration { get; set; }
    protected float Current { get; set; }
    protected float Previous { get; set; }
    protected float Step { get; set; }

    public float Elapsed
    {
        get
        {
            return Duration - Current;
        }

        set
        {
            Current = Mathf.Max(0, Duration - value);
        }
    }

    public bool Done
    {
        get
        {
            return Current <= 0;
        }
    }

    public float PercentDone
    {
        get
        {
            if (Duration > 0)
            {
                return Elapsed / Duration;
            }

            // A durationless timer will always be complete by convention.
            return 1.0f;
        }
    }

    public float PercentStep
    {
        get
        {
            // Will normally come back 0 or 1, except in the case where the timer just finished.
            if (Step >= 0)
            {
                return (Previous - Current) / Step;
            }

            return 0.0f;
        }
    }

    public Timer()
    {
        Duration = 0;
        Reset();
    }

    public Timer(float duration)
    {
        Duration = duration;
        Reset();
    }

    public void Reset()
    {
        Current = Duration;
        Previous = Current;
        Step = 0;
    }

    public void Reset(float duration)
    {
        Duration = duration;
        Reset();
    }

    public bool Tick(float amount)
    {
        Previous = Current;
        Step = amount;
        Current = Mathf.Max(0, Current - Step);

        return !Done;
    }

    public void Finish()
    {
        Previous = Current;
        Step = Current;
        Current = 0;
    }
}
