﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEditor;

public class Brick : MonoBehaviour
{

    [SerializeField]private GameObject desPrefab;

    public enum StateMachine
    {
        Free,
        Held,
        Thrown
    }

    public bool test = false;

    protected StateMachine _State = StateMachine.Free;
    public StateMachine State
    {
        get
        {
            return _State;
        }
        set
        {
            if (value != _State)
            {
                FinishState(_State);
                _State = value;
                InitState(_State);
            }
        }
    }

    public string PhaseLayer = "BrickPhased";
    public string DefaultLayer = "Brick";

    [SerializeField] private BrickSound bs;
    [SerializeField] private float bounciness = 1;
    [SerializeField] private float friction = 0;

    [SerializeField] private float maxWidth = 1;
    [SerializeField] private float maxHeight = 1;

    [SerializeField] private int maxHealth = 2;
    public bool ConvertScale = true;

    [SerializeField] private SpriteRenderer sr = null;

    private int currentHealth = 1;

    public Timer ThrownPhaseTimer = new Timer(.375f);

    private PhysicsMaterial2D bouncyPhyiscsMaterial;

    protected Rigidbody2D _Rigid = null;
    public Rigidbody2D Rigid
    {
        get
        {
            if (_Rigid == null)
                _Rigid = GetComponent<Rigidbody2D>();
            return _Rigid;
        }
    }

    protected PlayerControl Holder { get; set; }
    protected float GravityRestore { get; set; }
    protected float RotationStore { get; set; }
    protected float AimStore { get; set; }

    // Start is called before the first frame update
    void Start()
    {
        if (Rigid == null)
            Debug.LogError("brick.cs defaults are not set");

        if (sr == null)
            Debug.LogError("brick.cs defaults are not set");

        if (ConvertScale)
            AbsorbScale();

        bouncyPhyiscsMaterial = new PhysicsMaterial2D();
        bouncyPhyiscsMaterial.bounciness = bounciness;
        bouncyPhyiscsMaterial.friction = friction;
    }

    //only used for debug testing.
    private void Update()
    {
        if (test)
        {
            Debug.Log("brick.cs updated physics material");
            ChangeState();
            test = false;
        }
    }

    private void FixedUpdate()
    {
        RunState(State);

        if (transform.position.y < -5f)
            Destroy(gameObject);
    }


    private void OnCollisionEnter2D(Collision2D collision)
    {
        if(collision.gameObject.CompareTag("Brick"))
        {
            bs.PlayBrickOnBrickSound(Rigid.velocity.magnitude);
        }
    }

    public void AbsorbScale()
    {
        Vector3 scale = transform.localScale;
        transform.localScale = Vector3.one;

        SetSize(new Vector2(scale.x, scale.y));

    }

    public void OnSpawn()
    {
        ConvertScale = false;
        Randomize();
    }

    public void Randomize()
    {
        float x = Random.value * Random.value;
        float y = Random.value;
        float width = Mathf.Lerp(.8f, maxWidth, x);
        float height = Mathf.Lerp(.8f, maxHeight, y);

        SetSize(new Vector2(width, height));
        //transform.localScale = new Vector3(width, height, 0);

        currentHealth = Random.Range(1, maxHealth + 1);

        if (currentHealth >= 2)
            sr.color = Color.HSVToRGB(Random.Range(0.0f, 1.0f), 0.9f, 0.9f);
        else
            sr.color = Color.HSVToRGB(Random.Range(0.0f, 1.0f), 0.45f, 0.45f);
    }

    public void TakeDamage(int _damage)
    {
        currentHealth -= _damage;
        if (currentHealth < 1)
        {

            GameObject des = GameObject.Instantiate(desPrefab);
            Destroy(gameObject);
        }
        else
        {
            bs.PlayCrackSound();
            float h;
            float s;
            float v;
            Color.RGBToHSV(sr.color,out h,out s,out v);
            sr.color = Color.HSVToRGB(h, 0.45f, 0.45f);
        }
    }

    public void ChangeState()
    {
        Rigid.sharedMaterial = bouncyPhyiscsMaterial;
    }

    public void Grab(PlayerControl holder)
    {
        Holder = holder;
        State = StateMachine.Held;
    }

    public void SetSize(Vector2 size)
    {
        List<BoxCollider2D> seen = new List<BoxCollider2D>();
        foreach (var sprite in GetComponentsInChildren<SpriteRenderer>())
        {
            Vector3 scale = sprite.transform.localScale;
            Vector2 scale2d = new Vector2(scale.x, scale.y);
            var scaled = size / scale2d;

            var collider = sprite.GetComponentInParent<BoxCollider2D>();
            if (collider != null)
            {
                if (!seen.Contains(collider))
                {
                    Vector3 collide_scale = collider.transform.localScale;
                    Vector2 collide_scale2d = new Vector2(collide_scale.x, collide_scale.y);

                    Vector2 ratio = (collide_scale2d * collider.size) / (scale2d * sprite.size);

                    var collide_scaled = (size / collide_scale2d) * ratio;
                    collider.size = collide_scaled * ratio;
                    seen.Add(collider);
                }
            }

            sprite.size = scaled;
        }

        foreach (var collider in GetComponents<BoxCollider2D>())
        {
            if (seen.Contains(collider))
                continue;

            Vector3 scale = collider.transform.localScale;
            Vector2 scale2d = new Vector2(scale.x, scale.y);
            var scaled = size / scale2d;

            collider.size = scaled;
        }
    }

    public void Throw(Vector2 velocity)
    {
        Holder = null;
        State = StateMachine.Thrown;

        Rigid.AddForce(velocity);
    }

    public void Kick(Vector2 velocity)
    {
        Rigid.velocity = velocity;
    }

    protected void InitState(StateMachine state)
    {
        switch (state)
        {
            case StateMachine.Free:
                break;
            case StateMachine.Held:
                InitStateHeld();
                break;
            case StateMachine.Thrown:
                InitStateThrown();
                break;
        }
    }

    protected void RunState(StateMachine state)
    {
        switch (state)
        {
            case StateMachine.Free:
                break;
            case StateMachine.Held:
                RunStateHeld();
                break;
            case StateMachine.Thrown:
                RunStateThrown();
                break;
        }
    }

    protected void FinishState(StateMachine state)
    {
        switch (state)
        {
            case StateMachine.Free:
                break;
            case StateMachine.Held:
                FinishStateHeld();
                break;
            case StateMachine.Thrown:
                FinishStateThrown();
                break;
        }
    }

    protected void InitStateHeld()
    {

        bs.PlayPickupSound();
        State = StateMachine.Held;
        GravityRestore = Rigid.gravityScale;
        RotationStore = Rigid.rotation;

        Vector2 aim = Holder.AimVector;
        AimStore = Mathf.Atan2(aim.y, aim.x);
        Rigid.gravityScale = 0;

        foreach (var collider in GetComponents<Collider2D>())
        {
            collider.enabled = false;
        }
    }

    protected void RunStateHeld()
    {
        Vector2 aim = Holder.AimVector;
        float theta = Mathf.Atan2(aim.y, aim.x);

        Rigid.position = Holder.Rigid.position + Holder.MuzzleOffset;
        Rigid.velocity = Holder.Rigid.velocity;
        Rigid.rotation = RotationStore + ((theta - AimStore) * Mathf.Rad2Deg);
    }

    protected void FinishStateHeld()
    {
        Rigid.velocity = Rigid.velocity * .33f;
        Rigid.gravityScale = GravityRestore;
    }

    protected void InitStateThrown()
    {
        bs.PlayThrowSound();
        ThrownPhaseTimer.Reset();

        gameObject.layer = LayerMask.NameToLayer(PhaseLayer);

        foreach (var collider in GetComponents<Collider2D>())
        {
            collider.enabled = true;
        }
    }

    protected void RunStateThrown()
    {
        ThrownPhaseTimer.Tick(Time.fixedDeltaTime);

        if (ThrownPhaseTimer.Done)
            State = StateMachine.Free;
    }

    protected void FinishStateThrown()
    {
        gameObject.layer = LayerMask.NameToLayer(DefaultLayer);
        Rigid.gravityScale = GravityRestore;
    }
}
